/*
 *  Filename:  keys.c
 *
 *  Programmer:  Br. David Carlson
 *
 *  Date:  June 22, 1994
 *
 *  This is one of the source files for the ted editor project.  The
 *  functions in this file all have to do with the processing of special
 *  keys.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "screen.h"
#include "ted.h"


/* Global variables for quick access to commonly used items: */

extern LINE_TYPE *buffer;
extern MARK_RECORD mark;
extern char source_filename[80], TargetString[32];
extern INT8 mode, proceed, insert, modified, BlockMarked;
extern INT16 line_count, avail, first, last, current, file_row, clipboard,
   screen_row, screen_col, file_col, TargetLength;


/*  This function handles the pressing of the insert key. */

void handle_insert(void)
   {
   insert = 1 - insert;   /* reverse the value */
   display_color(NORMAL);
   display_color(WHITE);
   display_goto(1, 0);
   if (insert)
      printf("Insert  ");
   else
      printf("Typeover");
   display_goto(screen_row, screen_col);
   display_color(NORMAL);
   }


/*  The following function handles F3, the exit key. */

void handle_F3(void)
   {
   char reply;

   if (modified)
      {
      display_color(NORMAL);
      display_goto(11, 21);
      display_color(MAGENTA);
      printf(" Exit does not save your work.        ");
      display_goto(12, 21);
      printf(" Are you sure you want to exit (y/n)? ");
      display_goto(12, 58);
      reply = get_choice('y', 'n');

      if (reply == 'y')
         {
         display_color(NORMAL);
         proceed = FALSE;
         }
      else
         remove_window(11, 2, 21, 38);
      }
   else   /* file has not been modified since last saved */
      proceed = FALSE;
   }


/*  The following function handles F9, the save operation.  Note that if a
 *  block is marked it will ask whether you wish to save the block to a file.
 */

void handle_F9(void)
   {
   FILE *fp;
   char choice;
   INT8 ok;
   INT16 start_line, stop_line, index, len;
   char block_filename[36];

   if (BlockMarked)
      {
      display_color(NORMAL);
      display_goto(5, 22);
      display_color(MAGENTA);
      printf(" Save:                              ");
      display_goto(6, 22);
      printf(" b)   marked block to a new file    ");
      display_goto(7, 22);
      printf(" f)   entire file (under same name) ");
      display_goto(8, 22);
      printf(" Press b or f:                      ");
      display_goto(8, 37);
      choice = get_choice('b', 'f');
      }
   else
      choice = 'f';

   if (choice == 'b')
      {
      display_goto(5, 22);
      printf("                                    ");
      display_goto(6, 22);
      printf(" Enter filename to save block in:   ");
      display_goto(7, 22);
      printf("                                    ");
      display_goto(8, 22);
      printf("                                    ");
      len = get_string(block_filename, 34, 7, 23);
      block_filename[len] = NULL;
      remove_window(5, 4, 22, 36);

      fp = fopen(block_filename, "w");
      if (fp == NULL)
         {
         display_color(NORMAL);
         display_goto(11, 31);
         display_color(RED);
         printf(" Cannot open file ");
         display_goto(12, 31);
         printf(" Save aborted     ");
         display_goto(13, 31);
         printf(" Press ENTER      ");
         display_goto(13, 44);
         get_enter();
         remove_window(11, 3, 31, 18);
         return;
         }
      start_line = mark.first_buf_line;
      stop_line = mark.second_buf_line;
      }
   else   /* save entire file */
      {
      if (BlockMarked)
         remove_window(5, 4, 22, 36);
      fp = fopen(source_filename, "w");
      if (fp == NULL)
         {
         getch_done();
         display_color(NORMAL);
         display_clear();
         printf("\nCannot open file %s for writing\n", source_filename);
         exit(1);
         }
      start_line = first;
      stop_line = last;
      }

   display_color(NORMAL);
   display_goto(11, 32);
   display_color(MAGENTA);
   printf(" Saving ");

   if (choice == 'b')
      printf("block ");
   else
      printf("file  ");

   display_goto(12, 32);
   printf("              ");
   display_goto(13, 32);
   printf("              ");
   display_goto(12, 32);

   if ((ok = buffer_to_file(fp, start_line, stop_line)) == TRUE)
      printf(" ...completed ");
   else
      printf(" WRITE ERROR! ");

   fclose(fp);
   display_goto(13, 32);
   printf(" Press ENTER ");
   get_enter();

   if (choice == 'b')
      {
      unmark();
      index = get_index(2);
      display_text(2, 23, index);
      }
   else   /* saved entire file */
      {
      remove_window(11, 3, 32, 14);
      if (ok)
         modified = FALSE;
      }
   }


/*  This function handles a backspace key press. */

void handle_backspace(void)
   {
   register int k;
   INT16 prev, next, n, len, prev_len, total, top, count;
   LINE_TYPE *buf_ptr, *prev_ptr;

   if (BlockMarked)
      {
      beep();
      return;
      }

   buf_ptr = &buffer[current];

   if (screen_col == 0)
      {
      prev = buf_ptr->left;
      if (prev == NIL)
         beep();
      else
         {
         len = buf_ptr->length;
         prev_ptr = &buffer[prev];
         prev_len = prev_ptr->length;
         total = len + prev_len;

         if (total > 80)
            beep();
         else   /* backspace over ENTER mark, joining the 2 lines */
            {
            modified = TRUE;
            for (k = 0, n = prev_len; k < len; k++, n++)
               prev_ptr->info[n] = buf_ptr->info[k];
            prev_ptr->length = total;
            buf_ptr->length = 0;
            next = buf_ptr->right;
            prev_ptr->right = next;
            if (next == NIL)
               last = prev;
            else
               buffer[next].left = prev;

            /* Patch: 2/25/98 to return line to available list */
            buf_ptr->right = avail;
            avail = current;

            current = prev;
            file_row--;
            line_count--;
            if (prev_len == 80)
               screen_col = 79;
            else
               screen_col = prev_len;
            file_col = screen_col + 1;

            if (screen_row == 2)
               {
               top = current;
               count = 0;
               while (count < 21)
                  {
                  next = buffer[top].left;
                  if (next == NIL)
                     break;
                  count++;
                  top = next;
                  }
               screen_row += count;
               display_text(2, 23, top);
               }
            else
               {
               screen_row--;
               display_text(screen_row, 23, current);
               }

            show_row_col();
            display_goto(screen_row, screen_col);
            display_color(NORMAL);
            }
         }
      }
   else
      {
      len = buf_ptr->length;
      if (screen_col <= len)
         {
         modified = TRUE;
         len--;
         }
      file_col--;
      screen_col--;
      display_goto(screen_row, screen_col);

      for (k = screen_col; k < len; k++)
         {
         buf_ptr->info[k] = buf_ptr->info[k +1];
         putchar(buf_ptr->info[k]);
         }

      buf_ptr->info[len] = ' ';
      putchar(' ');
      show_file_col();
      buf_ptr->length = len;
      }
   }


/*  This function handles the DELETE key press. */

void handle_delete(void)
   {
   register int k;
   INT16 len;
   LINE_TYPE *buf_ptr;

   if (BlockMarked)
      {
      beep();
      return;
      }

   buf_ptr = &buffer[current];
   len = buf_ptr->length;

   if (screen_col < len)
      {
      modified = TRUE;
      len--;
      for (k = screen_col; k < len; k++)
         {
         buf_ptr->info[k] = buf_ptr->info[k + 1];
         putchar(buf_ptr->info[k]);
         }
      buf_ptr->info[len] = ' ';
      putchar(' ');
      display_goto(screen_row, screen_col);
      buf_ptr->length = len;
      }
   }


/*  This function handles the TAB key press. */

void handle_tab(void)
   {
   register int k;
   INT16 difference, column, len, new_len;
   LINE_TYPE *buf_ptr;

   column = screen_col / 3;
   column = 3 * column + 3;

   if (insert)
      {
      if (BlockMarked)
         {
         beep();
         return;
         }

      buf_ptr = &buffer[current];
      len = buf_ptr->length;
      difference = column - screen_col;
      new_len = len + difference;
      if ((new_len > 80) || (column > 79))
         beep();
      else
         {
         modified = TRUE;
         for (k = new_len - 1; k >= column; k--)
            buf_ptr->info[k] = buf_ptr->info[k - difference];
         for (k = column - 1; k >= screen_col; k--)
            buf_ptr->info[k] = ' ';
         for (k = screen_col; k < new_len; k++)
            putchar(buf_ptr->info[k]);
         screen_col = column;
         file_col = screen_col + 1;
         show_file_col();
         buf_ptr->length = new_len;
         }
      }
   else   /* typeover mode */
      {
      if (column > 79)
         beep();
      else
         {
         screen_col = column;
         file_col = column + 1;
         show_file_col();
         }
      }
   }


/*  This function handles the PageUp key press. */

void handle_page_up(void)
   {
   register int k;
   INT16 file_row_move, top_move, top;

   file_row_move = file_row - 1;
   if (file_row_move > 20)
      file_row_move = 20;
   top_move = file_row + 1 - screen_row;
   if (top_move > 20)
      top_move = 20;
   top = get_index(2);

   for (k = 0; k < top_move; k++)
      top = buffer[top].left;

   if (file_row_move == 0)
      {
      beep();
      file_row = 1;
      screen_row = 2;
      current = top;
      }
   else
      {
      if (top_move > 0)
         display_text(2, 23, top);

      for (k = 0; k < top_move; k++)
         current = buffer[current].left;

      if (top_move < file_row_move)
         {
         for (k = top_move; k < file_row_move; k++)
            current = buffer[current].left;
         screen_row = screen_row + top_move - file_row_move;
         }

      file_row = file_row - file_row_move;
      }

   show_file_row();
   }


/*  This function handles the PageDown key press. */

void handle_page_down(void)
   {
   register int k;
   INT16 back, difference, next, row, bottom, index, count;

   count = 0;
   row = screen_row;
   bottom = current;
   next = buffer[bottom].right;

   /* find row # and index of bottom line of data on screen */
   while ((row < 23) && (next != NIL))
      {
      bottom = next;
      next = buffer[bottom].right;
      row++;
      }

   /* count how many lines down the bottom line can be moved */
   if (next != NIL)
      {
      while ((count < 20) && (next != NIL))
         {
         count++;
         bottom = next;
         next = buffer[bottom].right;
         }
      }

   if (count == 0)   /* screen data doesn't move */
      {
      if (row == screen_row)   /* cursor doesn't move */
         beep();
      else   /* move the cursor down */
         {
         current = bottom;
         difference = row - screen_row;
         back = difference - 20;
         if (back > 0)
            {
            difference = 20;
            for ( ; back > 0; back--)
               current = buffer[current].left;
            }

         file_row = file_row + difference;
         screen_row = screen_row + difference;
         }
      }
   else   /* screen data does move */
      {
      index = get_index(2);   /* top index */
      for (k = 0; k < count; k++)
         {
         index = buffer[index].right;
         current = buffer[current].right;
         }

      display_text(2, 23, index);

      if (count < 20)
         {
         k = count;
         next = buffer[current].right;
         while ((next != NIL) && (k < 20))
            {
            current = next;
            next = buffer[current].right;
            k++;
            }
         file_row = file_row + k;
         screen_row = screen_row + k - count;
         }
      else
         file_row = file_row + count;
      }

   show_file_row();
   }


/*  The following helping function is used by handle_down_arrow and
 *  handle_enter to move down 1 line of the screen.
 */

void move_down(INT16 index)
   {
   register int row;

   current = index;
   file_row++;
   if (screen_row == 23)   /* scroll */
      {
      for (row = screen_row; row > 1; row--)
         {
         print_line(index, row, 0, 80);
         index = buffer[index].left;
         }
      }
   else
      screen_row++;
   }


/*  This function handles down arrow. */

void handle_down_arrow(void)
   {
   INT16 index;

   index = buffer[current].right;
   if (index == NIL)
      beep();
   else
      {
      move_down(index);
      show_file_row();
      }
   }


/*  This function handles up arrow. */

void handle_up_arrow(void)
   {
   INT16 index;

   index = buffer[current].left;
   if (index == NIL)
      beep();
   else
      {
      current = index;
      file_row--;
      if (screen_row == 2)   /* scroll */
         display_text(2, 23, index);
      else   /* no scroll */
         screen_row--;

      show_file_row();
      }
   }


/*  This function moves the cursor when left arrow is pressed. */

void handle_left(void)
   {
   if (screen_col == 0)
      beep();
   else
      {
      file_col--;
      screen_col--;
      show_file_col();
      }
   }


/*  This function moves the cursor when right arrow is pressed. */

void handle_right(void)
   {
   if (screen_col == 79)
      beep();
   else
      {
      file_col++;
      screen_col++;
      show_file_col();
      }
   }


/* This function moves the cursor when the home key is pressed. */

void handle_home(void)
   {
   file_col = 1;
   screen_col = 0;
   show_file_col();
   }


/* This function moves the cursor when the end key is pressed. */

void handle_end(void)
   {
   screen_col = buffer[current].length;
   if (screen_col == 80)
      {
      beep();
      screen_col = 79;
      }
   file_col = screen_col + 1;
   show_file_col();
   }


/*  The following handles the ENTER key press.  */

void handle_enter(void)
   {
   register int old_col, new_col, row;
   INT16 index, next;
   LINE_TYPE *buf_ptr, *current_ptr;

   next = buffer[current].right;

   if (insert)
      {
      if (BlockMarked)
         {
         beep();
         return;
         }

      if ((index = get_buffer_line()) == NIL)
         {
         allocation_error();
         return;
         }

      modified = TRUE;
      buf_ptr = &buffer[index];
      current_ptr = &buffer[current];
      current_ptr->right = index;
      buf_ptr->left = current;
      buf_ptr->right = next;
      if (next == NIL)
         last = index;
      else
         buffer[next].left = index;

      /* move any data from cursor to eoln to next line */
      new_col = 0;
      for (old_col = screen_col; old_col < current_ptr->length; old_col++)
         {
         buf_ptr->info[new_col++] = current_ptr->info[old_col];
         current_ptr->info[old_col] = ' ';
         }
      buf_ptr->length = new_col;
      if (screen_col < current_ptr->length)
         current_ptr->length = screen_col;

      if (screen_row == 23)   /* scroll */
         {
         for (row = 23; row > 1; row--)
            {
            print_line(index, row, 0, 80);
            index = buffer[index].left;
            }
         }
      else
         {
         display_text(screen_row, 23, current);
         screen_row++;
         }

      file_col = 1;
      screen_col = 0;
      current = current_ptr->right;
      line_count++;
      file_row++;

      show_row_col();
      display_goto(screen_row, screen_col);
      display_color(NORMAL);
      }
   else   /* typeover */
      {
      if (next == NIL)
         beep();   /* maybe change to insert a new line at bottom -- NOT */
      else
         {
         move_down(next);
         file_col = 1;
         screen_col = 0;
         show_row_col();
         display_goto(screen_row, screen_col);
         display_color(NORMAL);
         }
      }
   }


/*  The next function handles CTRL y; it deletes the current line.
 *  The line is also placed in the clipboard, so that it can be pasted
 *  back in later.
 */

void handle_ctrl_y(void)
   {
   INT16 index, prev, next, count, top;

   if (BlockMarked)
      {
      beep();
      return;
      }

   empty_clipboard();
   modified = TRUE;
   prev = buffer[current].left;
   next = buffer[current].right;
   line_count--;

   if ((prev == NIL) && (next == NIL))
      {
      line_count++;
      index = get_buffer_line();   /* should not be possible to fail */
      clipboard = current;
      current = first = last = index;
      buffer[index].left = NIL;
      buffer[index].right = NIL;
      screen_col = 0;
      file_col = 1;
      print_line(current, screen_row, 0, 80);
      show_file_col();
      }
   else if (next == NIL)
      {
      clipboard = current;
      buffer[clipboard].left = NIL;
      buffer[prev].right = NIL;
      last = current = prev;
      file_row--;
      if (screen_row > 2)   /* no scroll */
         {
         print_line(NIL, screen_row, 0, 80);
         screen_row--;
         }
      else   /* screen_row is 2, so scroll */
         {
         top = current;
         count = 0;
         while (count < 21)
            {
            next = buffer[top].left;
            if (next == NIL)
               break;
            count++;
            top = next;
            }
         screen_row += count;
         display_text(2, 23, top);
         }
      show_file_row();
      }
   else
      {
      clipboard = current;
      buffer[clipboard].left = NIL;
      buffer[clipboard].right = NIL;

      if (prev == NIL)
         first = next;
      else
         buffer[prev].right = next;

      buffer[next].left = prev;
      current = next;
      display_text(screen_row, 23, current);
      }
   }


/* The following function handles ctrl r by redrawing the screen. */

void handle_ctrl_r(void)
   {
   INT16 index;

   display_color(NORMAL);
   display_info_lines();
   index = get_index(2);
   display_text(2, 23, index);
   }


/*  The following function handles ctrl t by moving to the top of file. */

void handle_ctrl_t(void)
   {
   current = first;
   screen_row = 2;
   screen_col = 0;
   file_row = 1;
   file_col = 1;
   display_text(2, 23, current);
   show_row_col();
   display_goto(screen_row, screen_col);
   display_color(NORMAL);
   }


/*  The following function handles ctrl b by moving to the bottom of file. */

void handle_ctrl_b(void)
   {
   register int k;
   INT16 next;

   if (line_count < 23)   /* no scrolling needed */
      {
      next = buffer[current].right;
      while (next != NIL)
         {
         current = next;
         next = buffer[current].right;
         }
      screen_row = line_count + 1;
      screen_col = 0;
      file_row = line_count;
      file_col = 1;
      }
   else   /* scroll */
      {
      current = last;
      for (k = 23; k > 2; k--)
         current = buffer[current].left;
      screen_row = 23;
      screen_col = 0;
      file_col = 1;
      file_row = line_count;
      display_text(2, 23, current);
      current = last;
      }

   show_row_col();
   display_goto(screen_row, screen_col);
   display_color(NORMAL);
   }


/*  The following function handles ctrl e, deleting to end of line. */

void handle_ctrl_e(void)
   {
   LINE_TYPE *buf_ptr;
   INT16 len;
   register int k;

   if (BlockMarked)
      {
      beep();
      return;
      }

   buf_ptr = &buffer[current];
   len = buf_ptr->length;
   buf_ptr->length = screen_col;
   if (screen_col < len)
      modified = TRUE;
   for (k = screen_col; k < len; k++)
      buf_ptr->info[k] = ' ';

   print_line(current, screen_row, screen_col, len - screen_col);
   display_goto(screen_row, screen_col);
   }


/*  The following function searches the *StartLinePtr entry in the buffer to
 *  see if it can find a match for target that begins at *StartColPtr or 
 *  later.  The parameter len gives the length of the target string.
 *  Returns TRUE or FALSE.  If TRUE is returned, then *StartColPtr is the
 *  column index where the match begins.
 */

INT16 match(INT16 *StartLinePtr, INT16 *StartColPtr, INT16 len,
   char target[])
   {
   LINE_TYPE *buf_ptr;
   register int k;
   INT8 found, ok;
   INT16 current_len;
   char temp_string[80];

   found = FALSE;
   buf_ptr = &buffer[*StartLinePtr];
   current_len = buf_ptr->length;
   for (k = 0; k < current_len; k++)
      temp_string[k] = toupper(buf_ptr->info[k]);

   while ((! found) && (len + *StartColPtr <= current_len))
      {
      ok = TRUE;
      k = 0;
      while (ok && (k < len))
         {
         if (temp_string[*StartColPtr + k] != target[k])
            ok = FALSE;
         k++;
         }

      if (ok)
         found = TRUE;
      else
         (*StartColPtr)++;
      }

   if (! found)
      {
      *StartColPtr = 0;
      *StartLinePtr = buf_ptr->right;
      }

   return found;
   }


/*  The following function redisplays the screen to show the result of a 
 *  successful F2 search for a string.  The cursor is put toward the start of
 *  the string.  Note that the following global variables are changed:
 *  current, screen_row, screen_col, file_row, file_col.
 */

void MatchDisplay(INT16 start_line, INT16 start_col)
   {
   register int k;
   INT16 index, next;

   for (k = 0; current != start_line; k++)
      current = buffer[current].right;

   screen_col = start_col;
   file_col = start_col + 1;
   file_row += k;

   for (k = 15, index = current; k < 23; k++)
      {
      next = buffer[index].right;
      if (next == NIL)
         break;
      else
         index = next;
      }

   screen_row = 15 + 23 - k;

   for (k = screen_row, index = current; k > 2; k--)
      {
      next = buffer[index].left;
      if (next == NIL)
         break;
      else
         index = next;
      }

   screen_row = screen_row + 2 - k;
   display_text(2, 23, index);
   if (screen_col < 79)
      {
      screen_col++;
      file_col++;
      }

   show_row_col();
   display_goto(screen_row, screen_col);
   display_color(NORMAL);
   }


/*  The following function does most of the work for the string search
 *  used by F2 and F10.
 */

void search(void)
   {
   INT8 found;
   INT16 start_col, start_line;

   found = FALSE;
   start_col = screen_col;
   start_line = current;

   while ((! found) && (start_line != NIL))
      found = match(&start_line, &start_col, TargetLength, TargetString);

   if (found)
      MatchDisplay(start_line, start_col);
   else
      {
      beep();
      display_color(NORMAL);
      display_goto(15, 23);
      display_color(RED);
      printf(" That string could not be found. ");
      display_goto(16, 23);
      printf(" Press ENTER                     ");
      display_goto(16, 36);
      get_enter();
      remove_window(15, 2, 23, 33);
      }
   }


/*  The following function handles F2, the search key. */

void handle_F2(void)
   {
   register int k;

   display_color(NORMAL);
   display_goto(10, 9);
   display_color(MAGENTA);
   printf(" Enter string to search for:                                  ");
   display_goto(11, 9);
   printf(" Just press ENTER to skip.  After a search use F10 to repeat. ");
   TargetLength = get_string(TargetString, 32, 10, 38);
   remove_window(10, 2, 9, 62);
   for (k = 0; k < TargetLength; k++)
      TargetString[k] = toupper(TargetString[k]);

   if (TargetLength > 0)
      search();
   }


/*  The following function handles F10, the "search next" operation. */

void handle_F10(void)
   {
   if (TargetLength > 0)
      search();
   else
      beep();
   }


/*  The following function handles F4, the mark key.  The key should be used
 *  twice:  to mark both the start and the end of the desired block.  Pressing
 *  F4 a third times unmarks the currently marked block.
 */

void handle_F4(void)
   {
   register int k;
   INT16 temp, index;

   if (BlockMarked)
      {
      unmark();
      index = get_index(2);
      display_text(2, 23, index);
      }
   else if (mark.num_marks == 0)
      {
      mark.num_marks = 1;
      mark.first_buf_line = current;
      mark.first_file_row = file_row;
      }
   else if (mark.num_marks == 1)
      {
      BlockMarked = TRUE;
      mark.num_marks = 2;
      mark.second_buf_line = current;
      mark.second_file_row = file_row;

      if (mark.first_file_row > mark.second_file_row)
         {  /* swap info on first and second marks, so first <= second */
         temp = mark.first_buf_line;
         mark.first_buf_line = mark.second_buf_line;
         mark.second_buf_line = temp;
         temp = mark.first_file_row;
         mark.first_file_row = mark.second_file_row;
         mark.second_file_row = temp;
         }

      index = mark.first_buf_line;
      for (k = mark.first_file_row; k <= mark.second_file_row; k++)
         {
         buffer[index].marked = TRUE;
         index = buffer[index].right;
         }

      index = get_index(2);
      display_text(2, 23, index);
      }
   }


/*  The following function returns the lines on the clipboard list to the
 *  available list.
 */

void empty_clipboard(void)
   {
   register int index, next;

   index = clipboard;
   if (index == NIL)
      return;

   while ((next = buffer[index].right) != NIL)
      index = next;

   buffer[index].right = avail;
   avail = clipboard;
   clipboard = NIL;
   }


/*  The following function handles F5, the cut operation. */

void handle_F5(void)
   {
   INT16 prev, next, num_lines_cut, index, TargetScreenRow;

   if (! BlockMarked)
      {
      beep();
      return;
      }

   prev = buffer[mark.first_buf_line].left;
   next = buffer[mark.second_buf_line].right;

   if ((prev == NIL) && (next == NIL))
      {
      if ((index = get_buffer_line()) == NIL)
         {
         allocation_error();
         return;
         }
      }

   empty_clipboard();
   clipboard = mark.first_buf_line;
   buffer[clipboard].left = NIL;
   buffer[mark.second_buf_line].right = NIL;
   num_lines_cut = mark.second_file_row - mark.first_file_row + 1;
   line_count -= num_lines_cut;
   file_col = 1;
   screen_col = 0;

   if ((prev == NIL) && (next == NIL))
      {
      first = last = current = index;
      buffer[index].left = NIL;
      buffer[index].right = NIL;
      line_count = 1;
      file_row = 1;
      TargetScreenRow = 2;
      }
   else if (prev == NIL)
      {
      buffer[next].left = NIL;
      first = current = next;
      file_row = 1;
      TargetScreenRow = 2;
      }
   else if (next == NIL)
      {
      buffer[prev].right = NIL;
      last = current = prev;
      file_row = mark.first_file_row - 1;
      TargetScreenRow = 23;
      }
   else
      {
      buffer[prev].right = next;
      buffer[next].left = prev;
      current = prev;
      file_row = mark.first_file_row - 1;
      TargetScreenRow = 12;
      }

   display_text_adjusted(TargetScreenRow);
   unmark();
   modified = TRUE;
   }


/*  The following function handles F6, copying marked text to the clipboard.
 *  If the buffer doesn't have room, the operation is cancelled and an error
 *  message is presented.
 */

void handle_F6(void)
   {
   register int k;
   INT16 index, clip_index, prev_clip_index;
   LINE_TYPE *index_ptr, *clip_index_ptr;

   if (! BlockMarked)
      {
      beep();
      return;
      }

   empty_clipboard();
   index = mark.first_buf_line;

   if ((clip_index = get_buffer_line()) == NIL)
      {
      allocation_error();
      return;
      }

   clipboard = clip_index;
   prev_clip_index = clipboard;
   clip_index_ptr = &buffer[clip_index];
   index_ptr = &buffer[index];
   clip_index_ptr->left = NIL;
   clip_index_ptr->right = NIL;
   clip_index_ptr->length = index_ptr->length;

   for (k = 0; k < index_ptr->length; k++)
      clip_index_ptr->info[k] = index_ptr->info[k];

   while (index != mark.second_buf_line)
      {
      if ((clip_index = get_buffer_line()) == NIL)
         {
         allocation_error();
         empty_clipboard();
         return;
         }

      index = buffer[index].right;
      index_ptr = &buffer[index];
      clip_index_ptr = &buffer[clip_index];
      clip_index_ptr->left = prev_clip_index;
      clip_index_ptr->right = NIL;
      buffer[prev_clip_index].right = clip_index;
      clip_index_ptr->length = index_ptr->length;
      prev_clip_index = clip_index;

      for (k = 0; k < index_ptr->length; k++)
         clip_index_ptr->info[k] = index_ptr->info[k];
      }

   BlockMarked = FALSE;
   mark.num_marks = 0;
   unmark();
   index = get_index(2);
   display_text(2, 23, index);
   }


/*  The following function handles F7, pasting text in from the clipboard.
 *  Note that text is pasted in between the current line and the previous
 *  line.
 */

void handle_F7(void)
   {
   register int clip_end, next_line, count;
   INT16 prev;

   if ((BlockMarked) || (clipboard == NIL))
      {
      beep();
      return;
      }

   clip_end = clipboard;
   count = 1;

   while ((next_line = buffer[clip_end].right) != NIL)
      {
      clip_end = next_line;
      count++;
      }

   if ((prev = buffer[current].left) == NIL)
      first = clipboard;
   else
      buffer[prev].right = clipboard;

   buffer[clipboard].left = prev;
   buffer[current].left = clip_end;
   buffer[clip_end].right = current;

   current = clipboard;
   clipboard = NIL;
   line_count += count;
   file_col = 1;
   screen_col = 0;

   display_text_adjusted(screen_row);
   modified = TRUE;
   }


/*  The following function handles F8.  It retrieves the text from the
 *  file indicated by the user, inserting it just BEFORE the current line.
 */

void handle_F8(void)
   {
   FILE *fp;
   char filename[36];
   register int ch, col;
   INT16 row, index, prev, old_prev, new_col, first_line, len, next_line;
   INT8 last_was_newline, problem, empty_file;
   LINE_TYPE *buf_ptr;

   if (BlockMarked)
      {
      beep();
      return;
      }

   problem = NOERROR;
   display_color(MAGENTA);
   display_goto(5, 22);
   printf(" The retrieve operation merges some ");
   display_goto(6, 22);
   printf(" file into the current document.    ");
   display_goto(7, 22);
   printf(" Enter filename to be retrieved:    ");
   display_goto(8, 22);
   printf("                                    ");
   len = get_string(filename, 34, 8, 23);
   filename[len] = NULL;
   remove_window(5, 4, 22, 36);

   fp = fopen(filename, "r");
   if (fp == NULL)
      {
      report_error(OPENERROR);
      return;
      }

   row = 0;
   col = 0;
   if ((index = get_buffer_line()) == NIL)
      {
      report_error(ALLOCERROR);
      fclose(fp);
      return;
      }

   empty_file = TRUE;   /* until we find out otherwise */
   first_line = index;
   prev = index;
   old_prev = NIL;
   buf_ptr = &buffer[index];
   buf_ptr->left = NIL;
   buf_ptr->right = NIL;
   last_was_newline = FALSE;

   while (!feof(fp))
      {
      ch = fgetc(fp);
      if (ferror(fp))
         {
         problem = READERROR;
         break;
         }

      if (ch == '\n')
         {
         empty_file = FALSE;
         buf_ptr->length = col;
         /* must pad with blanks to end of standard line length */
         for ( ; col < 80; col++)
            buf_ptr->info[col] = ' ';
         row++;
         col = 0;
         if ((index = get_buffer_line()) == NIL)
            {
            problem = ALLOCERROR;
            break;
            }

         buf_ptr = &buffer[index];
         buffer[prev].right = index;
         buf_ptr->left = prev;
         buf_ptr->right = NIL;
         old_prev = prev;
         prev = index;
         last_was_newline = TRUE;
         }
      else if (ch == EOF)
         {
         if (last_was_newline)
            {
            return_buffer_line(index);
            index = old_prev;
            buffer[index].right = NIL;
            row--;
            }
         else if (empty_file)
            {
            fclose(fp);
            return;
            }
         else
            {   /* incomplete last line */
            buf_ptr->length = col;
            for ( ; col < 80; col++)
               buf_ptr->info[col] = ' ';
            }
         break;
         }
      else if (ch == '\t')
         {
         empty_file = FALSE;
         new_col = 3 * (col / 3) + 3;
         if (new_col > 80)
            {
            problem = LENGTHERROR;
            break;
            }
         for ( ; col < new_col; col++)
            buf_ptr->info[col] = ' ';
         last_was_newline = FALSE;
         }
      else   /* ch is an ordinary character */
         {
         empty_file = FALSE;
         if (col > 79)
            {
            problem = LENGTHERROR;
            break;
            }
         buf_ptr->info[col++] = ch;
         last_was_newline = FALSE;
         }
      }

   fclose(fp);
   if (problem == NOERROR)
      {
      modified = TRUE;
      if ((prev = buffer[current].left) == NIL)
         first = first_line;
      else
         buffer[prev].right = first_line;

      buffer[first_line].left = prev;
      buffer[current].left = index;
      buffer[index].right = current;

      current = first_line;
      line_count += row + 1;
      file_col = 1;
      screen_col = 0;
      display_text_adjusted(screen_row);
      }
   else
      {
      report_error(problem);
      index = first_line;
      while (index != NIL)
         {
         next_line = buffer[index].right;
         return_buffer_line(index);
         index = next_line;
         }
      }
   }


/*  The following function handles F1, the help screen key. */

void handle_F1(void)
   {
   INT8 screen_num;

   display_color(NORMAL);
   screen_num = 1;

   while (screen_num != 0)
      {
      display_goto(0, 0);
      help_screen(screen_num);
      display_goto(23, 79);
      get_screen_num(&screen_num);
      }

   handle_ctrl_r();
   }


/*  The following function displays the help screen given by screen_num. */

void help_screen(INT8 screen_num)
   {
   display_color(BLACK);
   if (screen_num != 3)
      printf("PgDn -- Next Help Screen");
   else
      printf("                        ");

   printf("      ");
   if (screen_num != 1)
      printf("PgUp -- Prev Help Screen");
   else
      printf("                        ");
   printf("      F3 -- Exit From Help\n");
   
   display_color(NORMAL);
   display_color(WHITE);
   printf("Help screen %1d for ted (text editor), vers", screen_num);
   printf("ion 1.3   (C) 1994 by Br. David Carlson\n");
   display_color(NORMAL);

   if (screen_num == 1)
      {
      printf("                                    SYNT");
      printf("AX:                                     \n");
      printf("ted filename      (edits a file of up to");
      printf(" 20 pages)                              \n");
      printf("   or                                   ");
      BLANK_END_OF_LINE;
      printf("ted filename -n   (edits a file no longe");
      printf("r than n pages)                         \n");
      printf("   (where 1 <= n <= 100, a page is 50 li");
      printf("nes, and a line holds 80 characters max)\n");
      print_blank_line();
      putchar('\n');
      printf("                                  BASIC ");
      printf("KEYS:                                   \n");
      printf("Arrow keys  Move 1 in direction of arrow");
      BLANK_END_OF_LINE;
      printf("PgUp  ----  Move up one screen in the fi");
      printf("le                                      \n");
      printf("PgDn  ----  Move down one screen in the ");
      printf("file                                    \n");
      printf("Enter  ---  Goto column 1 of next line (");
      printf("typeover); insert new line (insert mode)\n");
      printf("Backspace   Go left 1, deleting a charac");
      printf("ter (can backspace over an enter mark)  \n");
      printf("Insert  --  Toggle between insert mode a");
      printf("nd typeover mode                        \n");
      printf("Delete  --  Delete character at cursor  ");
      BLANK_END_OF_LINE;
      printf("Tab  -----  Move right to next tab stop ");
      printf("(typeover mode)                         \n");
      printf("            Insert blanks to next tab st");
      printf("op (insert mode)                        \n");
      printf("            Tab stops occur every 3 char");
      printf("acters                                  \n");
      print_blank_line();
      putchar('\n');
      printf("                              SAVING AND");
      printf(" EXITING:                               \n");
      printf("F9  ------  Save your file (under the sa");
      printf("me name)                                \n");
      printf("F3  ------  Exit from ted (does NOT save");
      printf(" your work)                             \n");
      print_blank_line();
      }
   else if (screen_num == 2)
      {
      printf("                              OTHER MOVE");
      printf("MENT KEYS:                              \n");
      printf("Home  ----  Move to first column of curr");
      printf("ent line                                \n");
      printf("End  -----  Move to last column of curre");
      printf("nt line                                 \n");
      printf("ctrl t  --  Move to top of file         ");
      BLANK_END_OF_LINE;
      printf("ctrl b  --  Move to bottom of file      ");
      BLANK_END_OF_LINE;
      print_blank_line();
      putchar('\n');
      printf("                              OTHER EDIT");
      printf("ING KEYS:                               \n");
      printf("ctrl y  --  Delete current line and plac");
      printf("e it in clipboard (F7 pastes it back in)\n");
      printf("ctrl e  --  Delete from cursor position ");
      printf("to end of line                          \n");
      print_blank_line();
      putchar('\n');
      printf("                              CUTTING AN");
      printf("D PASTING:                              \n");
      printf("ctrl y  --  Cuts out a single line and p");
      printf("laces it in the clipboard               \n");
      printf("F4  ------  Marks start and end of a blo");
      printf("ck of text (whole lines only)           \n");
      printf("            Pressing F4 a third time unm");
      printf("arks the block                          \n");
      printf("F5  ------  Cuts out the marked block an");
      printf("d places it in the clipboard            \n");
      printf("F6  ------  Places a copy of the marked ");
      printf("block in the clipboard                  \n");
      printf("F7  ------  Pastes the clipboard content");
      printf("s in just before the current line       \n");
      printf("F8  ------  Retrieves a file into your d");
      printf("ocument just before the current line    \n");
      printf("            (thus merging the two)      ");
      BLANK_END_OF_LINE;
      printf("F9  ------  When a block is marked, F9 g");
      printf("ives a choice of saving the entire file \n");
      printf("            or saving a copy of the mark");
      printf("ed block to a file which you can specify\n");
      print_blank_line();
      }
   else if (screen_num == 3)
      {
      printf("                                   SEARC");
      printf("HING:                                   \n");
      printf("F2  ------  Search for a string (from cu");
      printf("rsor onward)                            \n");
      printf("F10  -----  Search again for the same st");
      printf("ring (only valid after using F2)        \n");
      print_blank_line();
      putchar('\n');
      printf("                             REPAINTING ");
      printf("THE SCREEN:                             \n");
      printf("ctrl r  --  Repaints the whole screen --");
      printf(" useful if the screen gets confused     \n");
      print_blank_line();
      putchar('\n');
      printf("                        KNOWN LIMITATION");
      printf("S AND WARNINGS:                         \n");
      printf("The entire risk in using this software i");
      printf("s yours alone; no guarantees are made!  \n");
      printf("Max file size (using ted filename -100) ");
      printf("is 100 pages (5000 lines)               \n");
      printf("Max line length is 80 characters        ");
      BLANK_END_OF_LINE;
      printf("Do not press the ESC key (if you do, pre");
      printf("ss it a second time to cancel it)       \n");
      printf("Do not use the keys on the numeric keypa");
      printf("d on the right                          \n");
      printf("Do not use the alt key                  ");
      BLANK_END_OF_LINE;
      printf("Do not use any ctrl key combinations oth");
      printf("er than those listed                    \n");
      printf("If you use F10 to repeat a search for a ");
      printf("single character which happens to be in \n");
      printf("column 80, you will need to manually mov");
      printf("e the cursor to the start of the next   \n");
      printf("line before F10 will repeat the search  ");
      BLANK_END_OF_LINE;
      printf("Holding down cursor movement keys someti");
      printf("mes results in stray digits appearing in\n");
      printf("the document, so discrete key presses ar");
      printf("e to be preferred                       \n");
      printf("Note that you cannot do any editing whil");
      printf("e a block is marked                     \n");
      printf("Report any new problems to ID carlsond b");
      printf("y using the elm mailer (or similar)     ");
      }
   }


/*  The following function does all of the processing when a special
 *  key (such as a function key) has been pressed.
 */

void process_special(int key)
   {   /* incomplete */
   if (key == RIGHT_ARROW)
      handle_right();
   else if (key == LEFT_ARROW)
      handle_left();
   else if (key == DOWN_ARROW)
      handle_down_arrow();
   else if (key == UP_ARROW)
      handle_up_arrow();
   else if (key == PAGE_DOWN)
      handle_page_down();
   else if (key == PAGE_UP)
      handle_page_up();
   else if (key == HOME)
      handle_home();
   else if (key == END)
      handle_end();
   else if (key == DELETE)
      handle_delete();
   else if (key == BACKSPACE)
      handle_backspace();
   else if (key == TAB)
      handle_tab();
   else if (key == ENTER)
      handle_enter();
   else if (key == CTRL_B)
      handle_ctrl_b();
   else if (key == CTRL_Y)
      handle_ctrl_y();
   else if (key == CTRL_R)
      handle_ctrl_r();
   else if (key == CTRL_T)
      handle_ctrl_t();
   else if (key == CTRL_E)
      handle_ctrl_e();
   else if (key == F1)
      handle_F1();
   else if (key == F2)
      handle_F2();
   else if (key == F3)
      handle_F3();
   else if (key == F4)
      handle_F4();
   else if (key == F5)
      handle_F5();
   else if (key == F6)
      handle_F6();
   else if (key == F7)
      handle_F7();
   else if (key == F8)
      handle_F8();
   else if (key == F9)
     handle_F9();
   else if (key == F10)
      handle_F10();
   else if (key == INSERT)
      handle_insert();
   else
      beep();
   }

