/*
 *  Filename:  screen.h
 *
 *  Programmer:  Br. David Carlson
 *
 *  Reference:  Joy of C, by Miller & Quilici
 *
 *  Date:  May 14, 1994
 *
 *  This header file is to accompany SCREEN.C, which provides screen
 *  manipulation functions and immediate char input functions.
 */

/*  Macro display(color) is used with one of the 4 color constants
 *  shown below to change the color of output to the screen.
 */

#define display_color(color)  printf(color)

/*  Macro getch() is used to get an immediate key press without
 *  echoing the character to the screen.  You must call getch_init()
 *  (from screen.c) before using this macro.
 */

#define getch()   getchar()

/*  Macro beep() simply beeps the speaker. */

#define beep()    putchar(7)

#define TRUE      1
#define FALSE     0

/*  The following are colors to use for the display_color() macro. */

#define WHITE     "[7m"
#define GREEN     "[1;7m"
#define CYAN      "[1;4;7m"
#define NORMAL    "[0m"
#define BLACK     "[1m"
#define RED       "[1;4m"
#define BROWN     "[4;7m"
#define MAGENTA   "[4m"
#define BLINK     "[5m"
#define BLACK_BLINK "[4;7m"

/*  The following constants can be used to identify the keys read by
 *  the getkey() function provided by screen.c
 */

#define CTRL_B                 2
#define TAB                    9
#define ENTER                 10
#define CTRL_E                 5
#define CTRL_R                18
#define CTRL_T                20
#define CTRL_Y                25
#define ESC                   27
#define SPACE                 32
#define EXCLAMATION           33
#define DOUBLE_QUOTE          34
#define POUND                 35
#define DOLLAR                36
#define PERCENT               37
#define AMPERSAND             38
#define SINGLE_QUOTE          39
#define LEFT_PAREN            40
#define RIGHT_PAREN           41
#define TIMES                 42
#define PLUS                  43
#define COMMA                 44
#define MINUS                 45
#define PERIOD                46
#define SLASH                 47
#define COLON                 58
#define SEMICOLON             59
#define LESS_THAN             60
#define EQUALS                61
#define GREATER_THAN          62
#define QUESTION_MARK         63
#define AT_SIGN               64
#define LEFT_SQUARE_BRACKET   91
#define BACKSLASH             92
#define RIGHT_SQUARE_BRACKET  93
#define CARET                 94
#define UNDERSCORE            95
#define BACKQUOTE             96
#define LEFT_CURLY_BRACE     123
#define VERTICAL_BAR         124
#define RIGHT_CURLY_BRACE    125
#define TILDE                126
#define BACKSPACE            127
#define DOWN_ARROW           258
#define UP_ARROW             259
#define LEFT_ARROW           260
#define RIGHT_ARROW          261
#define HOME                 262
#define F1                   265
#define F2                   266
#define F3                   267
#define F4                   268
#define F5                   269
#define F6                   270
#define F7                   271
#define F8                   272
#define F9                   273
#define F10                  274
#define DELETE               330
#define INSERT               331
#define PAGE_DOWN            338
#define PAGE_UP              339
#define END                  347

/*  The following constants can be used to identify the type of key
 *  read by getkey:
 */

#define PRINTABLE_KEY          0
#define SPECIAL_KEY            1
#define UNKNOWN_KEY            2

